﻿using System;
using UnityEngine;
using UnityEngine.AI;
using Zenject;
using System.Collections;

public interface IMovement : IState, IOnCollision, IMove, IRotation, IClampAngle, IGround, IJump, IAnimatorMovement, IWaitForInteraction, IFixedTickable, ITickable, IInitializable
{

}

public interface IWaitForInteraction
{
	IEnumerator WaitForInteraction();
}

public interface ISceneTransition
{
	void SceneTransition();
}

public interface IMove
{
	void Move();
}

public interface IRotation
{
    void Rotate();
}

public interface IClampAngle
{
    void ClampAngle();
}

public interface IGround
{
	void GroundCheck();
}

public interface IJump
{
	void Jump();
}

public interface IState
{
	void Kill();
	void Revive();
}

public interface IOnCollision
{
	void OnCollisionStay(Collision other);
	void OnCollisionExit(Collision other);
}

[Serializable]
public class Movement : ScriptableObject, IMovement
{
    [Inject]
    public Settings settings;
    [Inject]
    public readonly Pawn.Settings pawn;
    [Inject]
    public readonly PlayerInputState inputState;

    public virtual void Initialize()
	{
		settings.rb = pawn.parentRoot.GetComponent<Rigidbody> ();
    }

	public virtual void Tick()
	{

	}   	

	public virtual void FixedTick()
	{

	}    

	public virtual void Kill()
	{

	}		

	public virtual void Revive()
	{

	}		

	public virtual void Jump()
	{

	}	

	public virtual void OnCollisionStay(Collision other)
	{
		if (other.gameObject.CompareTag("Platform") && settings.Grounded)
		{
			pawn.parentRoot.parent = other.transform;
		}
	}	

	public virtual void OnCollisionExit(Collision other)
	{
		pawn.parentRoot.SetParent(null);
	}

	public virtual void SceneTransition()
	{
		GroundCheck ();
		if (settings.Grounded) 
		{
			if (SceneController.instance.isFading) 
			{
				pawn.parentRoot.transform.position = new Vector3 (pawn.parentRoot.transform.position.x, pawn.parentRoot.transform.position.y, pawn.parentRoot.transform.position.z);	
			}
		}
	}

    public virtual void Move()
    {
    }    

	public virtual void Rotate()
	{
    }

	public virtual void ClampAngle()
	{
		if (inputState.WorldMovementInput != Vector3.zero)
		{
			inputState.LastFrameInput = inputState.WorldMovementInput;
		}
		else
		{
			return;
		}

		settings.lookDirection = Vector3.ProjectOnPlane(inputState.LastFrameInput, settings.GroundNormal);
		settings.groundQuaternion = Quaternion.LookRotation(settings.lookDirection, settings.GroundNormal);

		float groundEulerY = settings.groundQuaternion.eulerAngles.y;

		if (settings.groundQuaternion.eulerAngles.x < 180)
		{
			settings.groundQuaternion.eulerAngles = new Vector3(Mathf.Clamp(settings.groundQuaternion.eulerAngles.x, settings.clampAngle * -1, settings.clampAngle), settings.groundQuaternion.eulerAngles.y, settings.groundQuaternion.eulerAngles.z);
		}
		else if (settings.groundQuaternion.eulerAngles.x > 180)
		{
			settings.groundQuaternion.eulerAngles = new Vector3(Mathf.Clamp(settings.groundQuaternion.eulerAngles.x, 360 - settings.clampAngle, 360) - 360, settings.groundQuaternion.eulerAngles.y, settings.groundQuaternion.eulerAngles.z);
		}

		if (settings.groundQuaternion.eulerAngles.z < 180)
		{
			settings.groundQuaternion.eulerAngles = new Vector3(settings.groundQuaternion.eulerAngles.x, settings.groundQuaternion.eulerAngles.y, Mathf.Clamp(settings.groundQuaternion.eulerAngles.x, settings.clampAngle * -1, settings.clampAngle));
		}
		else if (settings.groundQuaternion.eulerAngles.z > 180)
		{
			settings.groundQuaternion.eulerAngles = new Vector3(settings.groundQuaternion.eulerAngles.x, settings.groundQuaternion.eulerAngles.y, Mathf.Clamp(settings.groundQuaternion.eulerAngles.z, 360 - settings.clampAngle, 360) - 360);
		}
	}

	public virtual void GroundCheck()
    {
		Debug.DrawRay(pawn.parentRoot.position + Vector3.back * .2f + (Vector3.up * settings.yGroundCheckOffset), Vector3.down * settings.groundCheckDistance);
		Debug.DrawRay(pawn.parentRoot.position + Vector3.forward * .2f + (Vector3.up * settings.yGroundCheckOffset), Vector3.down * settings.groundCheckDistance);

		RaycastHit hitInfo1, hitInfo2, hitInfo3;
		if (Physics.Raycast(pawn.parentRoot.position + Vector3.back * .2f   + (Vector3.up * settings.yGroundCheckOffset), Vector3.down, out hitInfo1, settings.groundCheckDistance, settings.groundMask, QueryTriggerInteraction.Ignore))
		{
			settings.GroundNormal = hitInfo1.normal;
			settings.GroundHit = hitInfo1.point;
			settings.Grounded = true;
		//	if(settings.currentGOTime > 0){
		//		settings.currentGOTime = 0f;
		//	}
		} else if (Physics.Raycast(pawn.parentRoot.position + Vector3.forward * .2f + (Vector3.up * settings.yGroundCheckOffset), Vector3.down, out hitInfo2, settings.groundCheckDistance, settings.groundMask, QueryTriggerInteraction.Ignore))
		{
			settings.GroundNormal = hitInfo2.normal;
			settings.GroundHit = hitInfo2.point;
			settings.Grounded = true;
		//	if(settings.currentGOTime > 0){
		//		settings.currentGOTime = 0f;
		//	}
		} else if (Physics.Raycast(pawn.parentRoot.position + (Vector3.up * settings.yGroundCheckOffset), Vector3.down, out hitInfo3, settings.groundCheckDistance, settings.groundMask, QueryTriggerInteraction.Ignore))
		{
			settings.GroundNormal = hitInfo3.normal;
			settings.GroundHit = hitInfo3.point;
			settings.Grounded = true;
		//	if(settings.currentGOTime > 0){
		//		settings.currentGOTime = 0f;
		//	}
		} else
		{
		//	settings.currentGOTime += Time.deltaTime;
		//	if(settings.currentGOTime >= settings.groundedOffTimer || settings.jumpedground){
				settings.GroundNormal = Vector3.up;
				settings.Grounded = false;
		//		settings.currentGOTime = 0f;
		//		settings.jumpedground = false;
		//	}

		}
	}

    public virtual void UpateMovement()
    {

    }

	public virtual IEnumerator WaitForInteraction ()
	{
		yield return null;
	}

    [Serializable]
    public class Settings
    {
		[Header("Interaction")]
		[HideInInspector]
		public Rigidbody rb;
		[HideInInspector]
		public bool handleInput;
		public WaitForSeconds inputHoldWait;

		[Header("Movement")]
        public float gravity;
        [HideInInspector]
        public float lastFrameVelocity;
        public float currentVelocity;
        public float maxVelocity;
        public float velocityMultiplier;
        [HideInInspector]
        public float angleMultiplier = 1;
        public int yFollowScale;
		public float acceleration;
		public float drag;

		[Header("Jump")]
		public int jumpForce;

		[Header("Rotation")]
		[HideInInspector]
		public Vector3 lookDirection; 
		public float turnScale;
		public int clampAngle;

		[Header("Ground")]
		public LayerMask groundMask;
		public float yGroundCheckOffset, groundCheckDistance, groundedOffTimer, currentGOTime;
		public bool platform;
		public bool Grounded;
		public bool Collision;
		public bool jumpedground;
        [HideInInspector]
        public Vector3 GroundHit;
        [HideInInspector]
        public Vector3 GroundNormal;
        [HideInInspector]
        public Quaternion groundQuaternion;

		[Header("Follow")]
		public bool followPlayer;
		public Transform followTarget;
        public byte startFollowDistance;
		public byte stopDistance;
		[HideInInspector]
		public Vector3 currentFollowPosition;
		[HideInInspector]
        public NavMeshAgent agent;

		public Settings(Settings settings)
		{
			Collision = settings.Collision;
			jumpForce = settings.jumpForce;
			gravity = settings.gravity;
			currentVelocity = settings.currentVelocity;
			maxVelocity = settings.maxVelocity;
			velocityMultiplier = settings.velocityMultiplier;
			acceleration = settings.acceleration;
			drag = settings.drag;
			yFollowScale = settings.yFollowScale;
			turnScale = settings.turnScale;
			clampAngle = settings.clampAngle;
			groundMask = settings.groundMask;
			yGroundCheckOffset = settings.yGroundCheckOffset;
			groundCheckDistance = settings.groundCheckDistance;
			Grounded = settings.Grounded;
			followPlayer = settings.followPlayer;
			startFollowDistance = settings.startFollowDistance;
			stopDistance = settings.stopDistance;
		}
    }
}